from .controller import Controller
from .connection import Connection, discover_devices, force_device_disconnect
from .connection import ConnectionException
from .feature import Feature, FeatureStatus, NotImplementedException
from .features.clean_filter import CleanFilterIndicator, CleanFilterIndicatorStatus, ResetCleanFilterTimer, ResetCleanFilterTimerStatus
from .features.fanspeed import FanSpeed, FanSpeedStatus, FanSpeedEnum
from .features.operationmode import OperationMode, OperationModeStatus, OperationModeEnum
from .features.power import PowerState, PowerStateStatus
from .features.setpoint import SetPoint, SetPointStatus
from .features.temperatures import Temperatures, TemperaturesStatus
from .cli import cli



"""This module contains the classes used to control the Operation Mode feature
"""
from enum import Enum
from typing import Dict
from pymadoka.feature import Feature, FeatureStatus
from pymadoka.connection import Connection

class OperationModeEnum(Enum):
     FAN = 0
     DRY = 1
     AUTO = 2
     COOL = 3
     HEAT = 4
     VENTILATION = 5
     def __str__(self):
        return self.name


class OperationModeStatus(FeatureStatus):

    """
    This class is used to store the Operation Mode status.
    
    Attributes:
        operation_mode (OperationModeEnum): Operation mode
       
    """

    DATA_IDX = 0x20
    
    def __init__(self,value:OperationModeEnum):
        """Inits with the operation mode.
        
        Attributes:
            operation_mode (OperationModeEnum): Operation mode
        """
        self.operation_mode = value
    
    def set_values(self, values:Dict[str,bytearray]):
        """See base class."""
        self.operation_mode = OperationModeEnum(int.from_bytes(values[self.DATA_IDX],"big"))
        
    def get_values(self) -> Dict[str,bytearray]:
        """See base class."""
        values = {}
        values[self.DATA_IDX] = self.operation_mode.value.to_bytes(1,"big")
        return values

class OperationMode(Feature):
    """
    This class is used to control the Operation Mode.

    Attributes:
        status (OperationModeStatus): Current status
    """
    def __init__(self, connection: Connection):
        """See base class."""
        self.status = None
        super().__init__(connection)

    def query_cmd_id(self) -> int:
        """See base class."""
        return 48
    
    def update_cmd_id(self) -> int:
        """See base class."""
        return 16432

    def new_status(self) -> FeatureStatus:
        """See base class."""
        return OperationModeStatus(OperationModeEnum.AUTO)



"""This module contains the classes used to control the Power feature (turn HVAC on/off)
"""

from typing import Dict
from pymadoka.feature import Feature, FeatureStatus
from pymadoka.connection import Connection

class PowerStateStatus(FeatureStatus):

    """
    This class is used to store the Power State status.
    
    Attributes:
        turn_on (bool): True if the HVAC is turned on, False otherwise
       
    """

    DATA_IDX = 0x20
    
    def __init__(self,turn_on:bool):
        """Inits with the power state.
        
        Attributes:
           turn_on (bool): True if the HVAC is turned on, False otherwise
        """
        self.turn_on = turn_on
    
    def set_values(self, values:Dict[str,bytearray]):
        """See base class."""
        self.turn_on = values[self.DATA_IDX][0] == 0x01
        
    def get_values(self) -> Dict[str,bytearray]:
        """See base class."""
        values = {}
        values[self.DATA_IDX] = bytes([0x01]) if self.turn_on else bytes([0x00])
        return values

class PowerState(Feature):

    """
    This class is used to control the HVAC Power (turn on/off)

    Attributes:
        status (PowerStateStatus): Current status
    """
    def __init__(self, connection: Connection):
        """See base class."""
        self.status = None
        super().__init__(connection)

    def query_cmd_id(self) -> int:
        """See base class."""
        return 32
    
    def update_cmd_id(self) -> int:
        """See base class."""
        return 16416

    def new_status(self) -> FeatureStatus:
        """See base class."""
        return PowerStateStatus(False)



"""This module contains the classes used to control the Set Point feature (temperatures set by the user)
"""

from typing import Dict
from pymadoka.feature import Feature, FeatureStatus
from pymadoka.connection import Connection

class SetPointStatus(FeatureStatus):
    """
    This class is used to store the Set Point temperatures.
    
    The values must be set as in Celsius degrees and are converted to the device format when read/written.

    No ranges validation is performed.

    Attributes:
        cooling_set_point (int): Cooling set point
        heating_set_point (int): Heating set point
    """

    COOLING_IDX = 0x20
    HEATING_IDX = 0x21

    def __init__(self,cooling_set_point:int, heating_set_point:int):
        """Inits the status with the set points
        
        Args: 
            cooling_set_point (int): Cooling set point
            heating_set_point (int): Heating set point
        """
        self.cooling_set_point = cooling_set_point
        self.heating_set_point = heating_set_point
   
    def set_values(self, values:Dict[str,bytearray]):
        """See base class."""
        self.cooling_set_point = round(int.from_bytes(values[self.COOLING_IDX],"big")/128.0)
        self.heating_set_point = round(int.from_bytes(values[self.HEATING_IDX],"big")/128.0)
        
    def get_values(self) -> Dict[str,bytearray]:
        """See base class."""
        values = {}
        values[self.COOLING_IDX] = (self.cooling_set_point*128).to_bytes(2,"big")
        values[self.HEATING_IDX] = (self.heating_set_point*128).to_bytes(2,"big")
        return values

class SetPoint(Feature):
    """
    This class is used to control the Set Point temperatures (temperatures set by the user)

    Attributes:
        status (SetPointStatus): Current status
    """
    def __init__(self, connection: Connection):
        """See base class."""
        self.status = None
        super().__init__(connection)

    def query_cmd_id(self) -> int:
        """See base class."""
        return 64
    
    def update_cmd_id(self) -> int:
        """See base class."""
        return 16448

    def new_status(self) -> FeatureStatus:
        """See base class."""
        return SetPointStatus(0,0)



"""This module contains the classes used to control the Fan Speed feature
"""

from enum import Enum
from typing import Dict
from pymadoka.feature import Feature, FeatureStatus
from pymadoka.connection import Connection

class FanSpeedEnum(Enum):
    HIGH = 5
    MID = 3
    LOW = 1
    AUTO = 0
    def __str__(self):
        return self.name

class FanSpeedStatus(FeatureStatus):

    """
    This class is used to store the Fan Speed status.
    
    Attributes:
        cooling_fan_speed (FanSpeedEnum): Cooling fan speed
        heating_fan_speed (FanSpeedEnum): Heating fan speed
       
    """

    COOLING_IDX = 0x20
    HEATING_IDX = 0x21

    def __init__(self,cooling_fan_speed:FanSpeedEnum, heating_fan_speed:FanSpeedEnum):
        """Inits with the cooling and heating fan speeds.

        Args:
            cooling_fan_speed (FanSpeedEnum): Cooling fan speed
            cooling_fan_speed (FanSpeedEnum): Heating fan speed
        """
        self.cooling_fan_speed = cooling_fan_speed
        self.heating_fan_speed = heating_fan_speed
   
    def set_values(self, values:Dict[str,bytearray]):
        """See base class."""
        cooling_value = int.from_bytes(values[self.COOLING_IDX],"big")
        heating_value = int.from_bytes(values[self.HEATING_IDX],"big")

        if cooling_value>=2 and cooling_value<=4:
            self.cooling_fan_speed = FanSpeedEnum.MID
        else:
            self.cooling_fan_speed = FanSpeedEnum(cooling_value)

        if heating_value>=2 and heating_value<=4:
            self.heating_fan_speed = FanSpeedEnum.MID
        else:
            self.heating_fan_speed = FanSpeedEnum(heating_value)
        
    def get_values(self) -> Dict[str,bytearray]:
        """See base class."""
        values = {}
        values[self.COOLING_IDX] = self.cooling_fan_speed.value.to_bytes(1,"big")
        values[self.HEATING_IDX] = self.heating_fan_speed.value.to_bytes(1,"big")
        return values

class FanSpeed(Feature):
    """
    This class is used to control the Fan Speed.

    Attributes:
        status (FanSpeedStatus): Current status
    """
    def __init__(self, connection: Connection):
        """See base class."""
        self.status = None
        super().__init__(connection)

    def query_cmd_id(self) -> int:
        """See base class."""
        return 80
    
    def update_cmd_id(self) -> int:
        """See base class."""
        return 16464

    def new_status(self) -> FeatureStatus:
        """See base class."""
        return FanSpeedStatus(FanSpeedEnum.AUTO,FanSpeedEnum.AUTO)



"""This module contains the classes used to control the Clean Filter feature (indicator status and reset indicator)
"""

from typing import Dict
from pymadoka.feature import Feature, FeatureStatus, NotImplementedException
from pymadoka.connection import Connection

class ResetCleanFilterTimerStatus(FeatureStatus):

    """
    This class is used to store the parameters used to reset the Clean Filter Indicator.
    """
    CLEAN_FILTER_TIMER_RESET_IDX = 0xFE

    def set_values(self, values):
        """See base class."""
        pass  

    def get_values(self):
        """See base class."""
        values = {}
        values[self.CLEAN_FILTER_TIMER_RESET_IDX] = bytes([0x01])
       
        return values


class ResetCleanFilterTimer(Feature):
    """
    This class is used to reset the Clean Filter Indicator.

    Attributes:
        status (ResetCleanFilterTimerStatus): Current status
    """

    def __init__(self, connection: Connection):
        """See base class."""
        self.status = None
        super().__init__(connection)

    def query_cmd_id(self) -> int:
        """See base class.
        
        Raises:
            NotImplementedException: This feature cannot be queried
        """
        raise NotImplementedException("This feature cannot be queried")
    
    def update_cmd_id(self)-> int:
        """See base class."""
        16928

    def new_status(self) -> FeatureStatus:
        """See base class."""
        return ResetCleanFilterTimerStatus()


class CleanFilterIndicatorStatus(FeatureStatus):

    """
    This class is used to store the Clean Filter Indicator status.
    
    Attributes:
        clean_filter_indicator (bool): True if indicator is turned on, False otherwise   
    """
    CLEAN_FILTER_IDX = 0x62
 
    def __init__(self,clean_filter_indicator: bool):
        """Inits with the clean filter indicator state.
        
        Attributes:
           clean_filter_indicator (bool): True if the indicator is turned on, False otherwise
        """
        self.clean_filter_indicator = clean_filter_indicator
   
    def set_values(self, values:Dict[int,bytearray]):
        """See base class."""
        self.clean_filter_indicator = values[self.CLEAN_FILTER_IDX][0] & 0x01 == 0x01
        
    def get_values(self) -> Dict[int,bytearray]:
        """See base class."""
        return {self.CLEAN_FILTER_IDX:bytes([0x00])}

class CleanFilterIndicator(Feature):

    """
    This class is used to retrieve the Clean Filter Indicator status.

    Attributes:
        status (CleanFilterIndicatorStatus): Current status
    """
    def __init__(self, connection: Connection):
        """See base class."""
        self.status = None
        super().__init__(connection)

    def query_cmd_id(self) -> int:
        """See base class."""
        return 256
    
    def update_cmd_id(self) -> int:
        """See base class.
        
        Raises:
            NotImplementedException: This feature cannot be updated
        """
        raise NotImplementedException("This feature cannot be updated")

    def new_status(self) -> FeatureStatus:
        """See base class."""
        return CleanFilterIndicatorStatus(False)



"""This module contains the device controller.
"""
import asyncio
import logging

from typing import Union, Dict
from enum import Enum
from pymadoka.feature import Feature, NotImplementedException
from pymadoka.connection import Connection, ConnectionException, ConnectionStatus
from pymadoka.features.fanspeed import FanSpeed
from pymadoka.features.operationmode import OperationMode
from pymadoka.features.power import PowerState
from pymadoka.features.setpoint import SetPoint
from pymadoka.features.temperatures import Temperatures
from pymadoka.features.clean_filter import CleanFilterIndicator,ResetCleanFilterTimer

logger = logging.getLogger(__name__)


class Controller:
    """This class implements the device controller.
    It stores all the features supported by the device and provides methods to operate globally on all the features.
    However, each feature can be queried/updated independently by accesing the feature attributes.

    Attributes:
        status (dict[string,FeatureStatus]): Last status collected from the features
        connection (Connection): Connection used to communicate with the device
        fan_speed (Feature): Feature used to control the fan speed
        operation_mode (Feature): Feature used to control the fan speed
        power_state (Feature): Feature used to control the fan speed
        set_point (Feature): Feature used to control the fan speed
        set_point (Feature): Feature used to control the fan speed
        clean_filter_indicator (Feature): Feature used to control the fan speed
    """
    def __init__(self, address: str, adapter: str = "hci0"):
        """Inits the controller with the device address.

        Args:
            address (str): MAC address of the device  
            adapter (str): Bluetooth adapter for the connection
        """


        if adapter is None:
            adapter = "hci0"
        
        self.status = {}
        self.connection = Connection(address,adapter = adapter)
        
        self.fan_speed = FanSpeed(self.connection)
        self.operation_mode = OperationMode(self.connection)
        self.power_state = PowerState(self.connection)
        self.set_point = SetPoint(self.connection)
        self.temperatures = Temperatures(self.connection)
        self.clean_filter_indicator = CleanFilterIndicator(self.connection)
        self.reset_clean_filter_timer = ResetCleanFilterTimer(self.connection)

    async def start(self):
        """Start the connection to the device.
        """        
        await self.connection.start()
       
    
    async def stop(self):
        """Stop the connection.
        """ 
        await self.connection.cleanup()
    
    async def update(self):
        """Iterate over all the features and query their status.
        """ 
        for var in vars(self).values():
            if isinstance(var,Feature): 
                try:
                    # Small delay to avoid DBUS errors produced when calls are too quick
                    await asyncio.sleep(0.3)
                    await var.query()
                except NotImplementedException as e:
                    pass
                except ConnectionAbortedError:
                    break
                except ConnectionException as e:
                    logger.debug(f"Connection error: {str(e)}")
                    pass
                except Exception as e:
                    logger.error(f"Failed to update {var.__class__.__name__}: {str(e)}")


    
    def refresh_status(self) -> Dict[str,Union[int,str,bool,dict,Enum]]:
        """Collect the status from all the features into a single status dictionary with basic types.

        Returns:
            dict[str,Union[int,str,bool,dict,Enum]]: Dictionary with the status of each feature represented with basic types
        """
        for k,v in vars(self).items():
            if isinstance(v,Feature): 
                if v.status is not None:
                    self.status[k] = vars(v.status)
            
        return self.status

    
    async def read_info(self) -> Dict[str,str]:
        """Reads the device info (Hardware revision, Software revision, Model, Manufacturer, etc)
        Returns:
            Dict[str,str]: Dictionary with the device info
        """
        
        return await self.connection.read_info()
       



    


"""This module encapsulates the data transfer protocol requirements defined by the device
"""

import logging
import math
import typing
from abc import ABC,abstractmethod


"""
Data transfer size 
"""
MAX_CHUNK_SIZE = 20

logger = logging.getLogger(__name__)

class TransportDelegate(ABC):
    """
    This interface defines the methods used by the Transport to notify the result of the rebuild process.
    """

    """Callback method used when a message has been sucessfully rebuilt.
         
    Args:
        data (bytearray): Rebuilt message data
    """
    @abstractmethod
    def response_rebuilt(self,data:bytearray):
        pass

    """Callback method used when the stored chunks of a message have been discarded.  
    """
    @abstractmethod
    def response_failed(self):
        pass


class Transport:

    """This class encapsulates the data packetizing required by the device (max 20 bytes per package as described in the protocol).

    Therefore, bigger data has to be split into smaller chunks and sent done in sequence.
    The implementation assumes we will receive all the chunks in order as they are being sent in serial order
    However, this will not work if more than one client is attached to the same device
         
    Attributes:
        delegate (`TransportDelegate`): The delegate to be notified when the a message has been rebuilt or discarded.
        chunks (List[bytearray]): Current rebuild chunks data
        last_id (int): ID of the last chunk stored in the list. This is used to check if newly received chunks follow the sequence
    """
    def __init__(self, delegate:TransportDelegate):
        """Inits the transport with the delegate.
        
        Args:
            delegate (`TransportDelegate`): The delegate to be notified when the a message has been rebuilt or discarded.
        """
        self.chunks:list(bytearray) = []
        self.delegate = delegate
        self.last_id = None
        

    def clear(self):
        """
        Clear the list of chunks
        """
        self.chunks.clear()


    def is_message_complete(self) -> bool:
        """Check if the stored chunks can be used to rebuild a message.

        Returns:
            bool: The return value. True for success, False otherwise.
        """
        if len(self.chunks) <= 0:
            return False
        
        expected_size = math.ceil(self.chunks[0][1] / MAX_CHUNK_SIZE)
       
        if len(self.chunks) != expected_size:
            return False

        return True

    def rebuild_chunk(self, chunk:bytearray):
        """Process the chunk.
         
            - The chunk has an id that follows the sequence of the stored chunks .If the message is complete, the delegate is notified
            - The chunk has an id that belongs to a different message .Current chunks are discarded and the delegate is notified

        Args:
            chunk (bytearray): The chunk data to be processed
        """
        if len(chunk) < 2:
            logger.info(F"Chunk received but discarded due to not enough data ({len(chunk)} bytes)")
            return
        
        chunk_id = chunk[0]
        
        if self.last_id is not None and chunk_id <= self.last_id:
            logger.debug("Chunks of a new message received while rebuilding another message. Discarding previous chunks...") 
            out = self.chunks_data()
            self.delegate.response_failed(out)


        self.last_id = chunk[0]
        
        self.chunks.append(chunk)
        
        if self.is_message_complete(): 

            logger.debug("Message complete. Processing...")
            out = self.chunks_data()
            self.last_id = None
            self.delegate.response_rebuilt(out)


    def chunks_data(self):
        out = bytearray()
        
        for c in self.chunks:
            out.extend(c[1:])
        self.chunks.clear()

        return out

    def split_in_chunks(self, data:bytearray) -> typing.List[bytearray]:
        """Split the data in MAX_CHUNK_SIZE bytes chunks. If more than one chunk is produced, numerate each chunk in the sequence.
        
        Args:
            data (bytearray): The data to be split
        Returns:
            List[bytearray]: List of chunks
        """
        chunks:list(bytearray) = []

        idx = 0
        while True:
            chunk = data[idx*19:min((idx+1)*19,len(data))]
            chunks.append(bytearray(idx.to_bytes(1,"big")) + chunk)
            idx+=1
            if idx*19 >= len(data):
                break

        return chunks
          




import asyncio
from asyncio.exceptions import CancelledError
from asyncio.futures import Future
import logging                                    
 
import subprocess
import sys
from enum import Enum

from bleak import BleakClient,BleakScanner,discover
from typing import Dict

from pymadoka.transport import Transport, TransportDelegate
from pymadoka.consts import NOTIFY_CHAR_UUID, WRITE_CHAR_UUID, SEND_MAX_TRIES

logger = logging.getLogger(__name__)

class ConnectionException(Exception):
    """Exceptions are documented in the same way as classes.

    The __init__ method may be documented in either the class level
    docstring, or as a docstring on the __init__ method itself.

    Either form is acceptable, but the two should not be mixed. Choose one
    convention to document the __init__ method and be consistent with it.

     """
    pass


class ConnectionStatus(Enum):
    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2
    ABORTED = 3

DISCOVERED_DEVICES_CACHE = []

async def discover_devices(timeout=5,adapter="hci0", force_disconnect = True):
    """Trigger a bluetooth devices discovery on the adapter for the timeout interval.
    
    This method must be called before any connection attempt.
    
    Args:
        timeout(int): Timeout used for the device discovery (default 5s)
        adapter (str): Bluetooth adapter
    """
    global DISCOVERED_DEVICES_CACHE

    scanner = BleakScanner(adapter = adapter)
    await scanner.start()
    await asyncio.sleep(timeout)
    await scanner.stop()
    DISCOVERED_DEVICES_CACHE = await scanner.get_discovered_devices()
    return DISCOVERED_DEVICES_CACHE

async def force_device_disconnect(address):

    """Force a device disconnect so it can be listed during the scan.

    This method relays on the `bluetoothctl` tool to disconnect the device.
    """

    logger.debug("Forcing disconnect...")
    process = await asyncio.create_subprocess_exec(
        "bluetoothctl","disconnect",address, 
        stdout=asyncio.subprocess.PIPE, 
        stderr=asyncio.subprocess.PIPE
        )
    # Wait for the subprocess to finish
    stdout, stderr = await process.communicate()
    # Progress
    if process.returncode != 0:
        logger.debug(f"Disconnect failed: {stderr.decode().strip()}")

class Connection(TransportDelegate):
    
    """Bluetooth client"""

    client: BleakClient = None

    """This class implements the bluetooth connection to the device.

    It communicates with the device to send and receive data that is passed to the `Transport` to be rebuilt.
    all the features supported by the device and provides methods to operate globally on all the features.
    However, each feature can be queried/updated independently by accesing the feature attributes.

    Attributes:
        client (BleakClient): Bluetooth device client
        transport (`Transport`): Transport used for the protocol
        address (str): MAC address of the device
        name (str): Name of the device when available
        current_future (Future): Future of the current command being processed
        connected (ConnectionStatus): Status of the connection
        adapter (str): Bluetooth adapter used for the client
        
    """

    def __init__(
        self,
        address: str,
        adapter: str
    ):
        """Inits the connection with the device address.

        Args:
            address (str): MAC address of the device
            adapter(str): Bluetooth adapter for the client
        """
        self.adapter = adapter
        self.address = address
        self.name = self.address
        self.connection_status = ConnectionStatus.DISCONNECTED
        self.last_info = None
        self.transport = Transport(self)
        self.current_future = None
        self.requests = {}
        
    def on_disconnect(self, client: BleakClient):
        self.connection_status = ConnectionStatus.DISCONNECTED
        # Put code here to handle what happens on disconnet.
        logger.info(f"Disconnected {self.address}!")
        asyncio.create_task(self.start())
        
    async def cleanup(self):
        if self.client:
            await self.client.stop_notify(NOTIFY_CHAR_UUID)
            await self.client.disconnect()
        self.connection_status = ConnectionStatus.DISCONNECTED

    async def start(self):
        """Starts the connection.
        
        Firstly, the device has to be registered in the Bluez Service so it becomes available via DBUS, so a device scan has to be performed.
        Once the devices are found, the client is registered and connected to the device that matches the device address.

        However, if the device is already connected, it requires to be disconnected so it can be found during the scan, as the device is not listed otherwise.

        Args:
            address (str): MAC address of the device
            force_disconnect (bool): Force a hard disconnect of the device. The device is usually disconnected to ensure a better communication (default True)
            device_discovery_timeout(int): Timeout used for the device discovery (default 5s)
        """
        logger.debug(F"Starting connection manager on {self.address}")
        self.connection_status = ConnectionStatus.CONNECTING
        while (not self.connection_status == ConnectionStatus.CONNECTED and 
               not self.connection_status == ConnectionStatus.ABORTED):
            try:
                if self.client:
                    await self._connect()
                else:
                    await self._select_device()
                await asyncio.sleep(2.0)       
            except ConnectionAbortedError as e:
                 raise e  
            except CancelledError as e:
                 logger.error(str(e))  
    async def _connect(self):
        try:
            await self.client.connect()
            connected = await self.client.is_connected()
            if connected:
                logger.info(F"Connected to {self.address}")
                self.client.set_disconnected_callback(self.on_disconnect)
                self.connection_status = ConnectionStatus.CONNECTED
                await self.client.start_notify(
                    NOTIFY_CHAR_UUID, self.notification_handler,
                )
        
            else:
               logger.info(f"Failed to connect to {self.address.name}")
        except Exception as e:
            if not "Software caused connection abort" in str(e):
                logger.error(e)
            logger.debug("Reconnecting...")

    async def _select_device(self):
        """Scan bluetooth devices searching for the thermostat address so it is registered in the DBUS and available to the client.
        """
        logger.debug("Bluetooh LE hardware warming up...")
    
        for d in DISCOVERED_DEVICES_CACHE:
            if d.address.upper() == self.address.upper(): 
                self.client = BleakClient(d, adapter = self.adapter)
                self.name = d.name
                break
        if self.client == None:
            self.connection_status = ConnectionStatus.ABORTED
            raise ConnectionAbortedError(f"Could not find bluetooth device for the address {self.address}. Please follow the instructions on device pairing.")

    def notification_handler(self, sender: str, data: bytearray):
        """This callback is used to receive the data read from the device (chunks) and attempt to rebuild the message.
        
        Args:
            sender (str) : Client ID
            data (bytearray): Data to be rebuilt
        """
        self.transport.rebuild_chunk(data)
       
    def cmd_id_to_bytes(self,cmd_id:int):
        return bytearray([0x00]) + cmd_id.to_bytes(2,"big")
    def bytes_to_cmd_id(self,data:bytes):
        return int.from_bytes(data[2:4],"big")

    async def send(self,cmd_id:int,data:bytearray):

        """This method is used to send data to the device.
        The `transport` is used to split the data into chunks as required by the communication protocol and these chunks are sent in order to the device.
        
        Args:
            cmd_id (str) : Command ID to be sent
            data (bytearray): Data to be sent
        Returns:
            Future: Callers of this methods must await this Future to receive the result of the command execution
        """

        cmd_response = asyncio.get_event_loop().create_future()
        if not cmd_id in self.requests:
            self.requests[cmd_id] = []

        self.requests[cmd_id].append(cmd_response)

        if self.connection_status is not ConnectionStatus.CONNECTED:
            cmd_response.cancel()
            return cmd_response
    
        # length, 0x00, cmdid, payload
        payload = bytearray([0x00]) + self.cmd_id_to_bytes(cmd_id) + data

        payload[0] = len(payload)

        logger.debug(f"Sending cmd payload: {payload}")
       
        chunks = self.transport.split_in_chunks(payload)
        sent = 0
        
        self.current_cmd_id = cmd_id
        for chunknum,chunk in enumerate(chunks):
            for i in range(0,SEND_MAX_TRIES):
                try:
                    if self.connection_status is not ConnectionStatus.CONNECTED:
                        cmd_response.cancel()
                        return cmd_response
                   
                    await self.client.write_gatt_char(WRITE_CHAR_UUID,chunk)
                    logger.debug(F"CMD {cmd_id}. Chunk #{chunknum+1}/{len(chunks)} sent with size {len(chunk)} bytes")
                    sent += 1
                    break
                except CancelledError as e:
                    logger.debug(F"Send command failed. Retrying ({i}/{SEND_MAX_TRIES}) for chunk #{chunknum} : {str(e)}", exc_info = e)
                    await asyncio.sleep(1)      
                except Exception as e:
                    logger.debug(F"Send command failed. Retrying ({i}/{SEND_MAX_TRIES}) for chunk #{chunknum} : {str(e)}")
                    await asyncio.sleep(1)    

        if sent != len(chunks) and self.connection_status == ConnectionStatus.CONNECTED:
            raise ConnectionException("Command chunks could not be sent")

        return cmd_response

    def response_rebuilt(self,data:bytearray):
        """This callback is used to receive messages rebuilt by the transport. 

        The messages are used to resolve the future used when the command was sent.

        See base class `TransportDelegate`."""

        if len(data)<=4:
            return

        cmd_id = self.bytes_to_cmd_id(data)

        if not cmd_id in self.requests:
            return
        if len(self.requests[cmd_id])>0:
            req = self.requests[cmd_id].pop(0)
            if req.done():
                return
            req.set_result(data)

    def response_failed(self,data:bytearray):
        """This callback is used to cancel the future used when the command was sent.

        See base class `TransportDelegate`."""
        
        if len(data)<=4:
            return

        cmd_id = self.bytes_to_cmd_id(data)

        if not cmd_id in self.requests:
            return

        if len(self.requests[cmd_id])>0:
            req = self.requests[cmd_id].pop(0)
            if req.done():
                return
            req.cancel()
     
    async def read_info(self) -> Dict[str,str]:
        """This method is used to retrieve the information stored in the Bluetooth Services available in the device. 

        This information is related to the Software Version, Hardware Version, Model Number and others.

        Returns:
            Dict[str,str]: Dictionary with all the info values
        """
        try:
            if self.last_info:
                 return self.last_info
            
            if self.connection_status is not  ConnectionStatus.CONNECTED:
                return {}

            values = {}

            for service in self.client.services:
                logger.debug("[Service] {0}: {1}".format(service.uuid, service.description))
                for char in service.characteristics:
                    if "read" in char.properties:
                        try:
                            raw = await self.client.read_gatt_char(char.uuid)
                            value = None
                            
                            try:
                                if char.description.endswith(" ID"): 
                                    value = raw.hex().replace("fe","-").replace("ff","")
                                else:
                                    value = raw.decode()
                            except:
                                value = str(raw)
                            values[char.description] = value
                            logger.debug(
                                "\t[Characteristic] {0}: (Handle: {1}) ({2}) | Name: {3}, Value: {4} ".format(
                                char.uuid,
                                char.handle,
                                ",".join(char.properties),
                                char.description,
                                value,
                            )
                        )
                        except Exception as e:
                            logger.error(e)


            self.last_info = values
            return self.last_info
        except Exception as e:
            logger.error(e)
            raise e
        
if __name__ == "__main__":

    async def main(connection:Connection):
        try:
            await discover_devices()
            start_task = await asyncio.create_task(connection.start())
            connection_status = await asyncio.gather(start_task)
            if connection_status.result() == ConnectionStatus.CONNECTED:
                await connection.client.write_gatt_char(WRITE_CHAR_UUID,[0x00,0x06,0x00,0x20,0x00,0x00])
        except Exception as e:
           logging.error(str(e))
           asyncio.get_event_loop().stop()
        
    logging.basicConfig(level=logging.NOTSET)
    # Create the event loop.
    loop = asyncio.get_event_loop()
    address = sys.argv[1]
    connection = Connection(address )
    try:
        asyncio.ensure_future(main(connection))
        loop.run_forever()
    except KeyboardInterrupt:
        logging.info()
        logging.info("User stopped program.")
   
    finally:
        
        logging.info("Disconnecting...")
        loop.run_until_complete(connection.cleanup())
    



"""This module contains the classes used to control the Temperature feature (temperatures read by the device)
"""

from typing import Dict
from pymadoka.feature import Feature, FeatureStatus, NotImplementedException
from pymadoka.connection import Connection

class TemperaturesStatus(FeatureStatus):


    INDOOR_IDX = 0x40
    OUTDOOR_IDX = 0x41

    """
    This class is used to store the indoor/outdoor temperatures.

    Attributes:
        indoor (int): Indoor temperature
        outdoor (int): Outdoor temperature
    """
    def __init__(self,indoor:int, outdoor:int):
        """Inits the feature with the indoor/outdoor temperatures.
        
        Args:
            indoor (int): Indoor temperature
            outdoor (int): Outdoor temperature
        """
        self.indoor = indoor
        self.outdoor = outdoor
   
    def set_values(self, values:Dict[str,bytearray]):
        """See base class."""
        self.indoor = values[self.INDOOR_IDX][0]
        self.outdoor = values[self.OUTDOOR_IDX][0]
        if self.outdoor == 0xff:
            self.outdoor = None;
        
    def get_values(self) -> Dict[str,bytearray]:
        """See base class."""
        values = {}
        values[self.INDOOR_IDX] = (self.indoor*128).to_bytes(2,"big")
        if self.outdoor is not None:
            values[self.OUTDOOR_IDX] = (self.outdoor*128).to_bytes(2,"big")
        return values

class Temperatures(Feature):

    """
    This class is used to retrieve the indoor/outdoor temperatures as registered by the device.

    Attributes:
        status (TemperaturesStatus): Current status
    """
    def __init__(self, connection: Connection):
        """See base class."""
        self.status = None
        super().__init__(connection)

    def query_cmd_id(self) -> int:
        """See base class."""
        return 272
    
    def update_cmd_id(self) -> int:
        """See base class. This method has not been implemented.

        Raises:
            NotImplementedException: This feature cannot be updated
        """
        raise NotImplementedException("This feature cannot be updated")

    def new_status(self) -> FeatureStatus:
        """See base class."""
        return TemperaturesStatus(0,0)



import asyncio
from asyncio.exceptions import CancelledError
import logging
import click
import json
import yaml

import paho.mqtt.client as mqtt
from paho.mqtt import MQTTException
from typing import Any,Dict
from functools import wraps
from pymadoka.connection import Connection, ConnectionStatus, ConnectionException, discover_devices, force_device_disconnect
from pymadoka.controller import Controller
from pymadoka.features.fanspeed import FanSpeedEnum, FanSpeedStatus
from pymadoka.features.setpoint import SetPointStatus
from pymadoka.features.operationmode import OperationModeStatus, OperationModeEnum
from pymadoka.features.power import PowerStateStatus
from pymadoka.features.clean_filter import ResetCleanFilterTimerStatus
from pymadoka import Feature

logger = logging.getLogger(__name__)

# Taken from paho-mqtt examples to integrate with asyncio loop

class AsyncioHelper:
    def __init__(self, loop, client):
        self.loop = loop
        self.client = client
        self.client.on_socket_open = self.on_socket_open
        self.client.on_socket_close = self.on_socket_close
        self.client.on_socket_register_write = self.on_socket_register_write
        self.client.on_socket_unregister_write = self.on_socket_unregister_write

    def on_socket_open(self, client, userdata, sock):
        def cb():
            client.loop_read()

        self.loop.add_reader(sock, cb)
        self.misc = self.loop.create_task(self.misc_loop())

    def on_socket_close(self, client, userdata, sock):
        self.loop.remove_reader(sock)
        self.misc.cancel()

    def on_socket_register_write(self, client, userdata, sock):

        def cb():
            client.loop_write()

        self.loop.add_writer(sock, cb)

    def on_socket_unregister_write(self, client, userdata, sock):
        self.loop.remove_writer(sock)

    async def misc_loop(self):
        while self.client.loop_misc() == mqtt.MQTT_ERR_SUCCESS:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break


async def set_operation_mode(controller:Controller,payload:str):
    """
    Callback used to set the operation mode. It will issue a turn on/off command 
    depending on the value of the mode (OFF if 'OFF', ON otherwise)
    Args:
        controller (`Controller`): Device controller
        payload (str): The payload will be converted to the values accepted by the controller
    """
    try:
        value = payload.decode("utf-8").upper()
        if value != "OFF":
            status = OperationModeStatus(OperationModeEnum[value])
            await controller.operation_mode.update(status)
        await controller.power_state.update(PowerStateStatus(value != "OFF"))
    except CancelledError as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except ConnectionException as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except ConnectionAbortedError as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except Exception as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    
async def set_fan_speed(controller:Controller,payload:str):
    """
    Callback used to set the fan speed
    Args:
        controller (`Controller`): Device controller
        payload (str): The payload will be converted to the values accepted by the controller
    """

    try:

        value = payload.decode("utf-8").upper()
        new_cooling_fan_speed = controller.fan_speed.status.cooling_fan_speed
        new_heating_fan_speed = controller.fan_speed.status.heating_fan_speed
        if (controller.operation_mode.status.operation_mode == OperationModeEnum.AUTO or
           controller.operation_mode.status.operation_mode == OperationModeEnum.DRY or
           controller.operation_mode.status.operation_mode == OperationModeEnum.FAN):
            new_cooling_fan_speed = value
            new_heating_fan_speed = value
        elif controller.operation_mode.status.operation_mode == OperationModeEnum.HEAT:
            new_heating_fan_speed = value
        elif controller.operation_mode.status.operation_mode == OperationModeEnum.COOL:
            new_cooling_fan_speed = value

        status = FanSpeedStatus(FanSpeedEnum[new_cooling_fan_speed],
                                FanSpeedEnum[new_heating_fan_speed])
        await controller.fan_speed.update(status)
    except CancelledError as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except ConnectionException as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except ConnectionAbortedError as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except Exception as e:
        logging.error(f"Could not update operation mode: {str(e)}")

async def set_power_state(controller:Controller,payload:str):
    """
    Callback used to set the power state
    Args:
        controller (`Controller`): Device controller
        payload (str): The payload will be converted to the values accepted by the controller
    """
    try:
        status = PowerStateStatus(payload=="on")
        await controller.power_state.update(status)
    except CancelledError as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except ConnectionException as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except ConnectionAbortedError as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except Exception as e:
        logging.error(f"Could not update operation mode: {str(e)}")

async def set_set_point_state(controller:Controller,payload:str):
    """
    Callback used to set the set point (target temperature)
    Args:
        controller (`Controller`): Device controller
        payload (str): The payload will be converted to the values accepted by the controller
    """
    try:
        value = int(payload)
        new_cooling_set_point = controller.set_point.status.cooling_set_point
        new_heating_set_point = controller.set_point.status.heating_set_point

        if (controller.operation_mode.status.operation_mode == OperationModeEnum.AUTO or
           controller.operation_mode.status.operation_mode == OperationModeEnum.DRY or
           controller.operation_mode.status.operation_mode == OperationModeEnum.FAN):
            new_cooling_set_point = value
            new_heating_set_point = value 
        elif controller.operation_mode.status.operation_mode == OperationModeEnum.HEAT:
            new_heating_set_point = value
        elif controller.operation_mode.status.operation_mode == OperationModeEnum.COOL:
            new_cooling_set_point = value
        await controller.set_point.update(SetPointStatus(new_cooling_set_point,new_heating_set_point))
    except CancelledError as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except ConnectionException as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except ConnectionAbortedError as e:
        logging.error(f"Could not update operation mode: {str(e)}")
    except Exception as e:
        logging.error(f"Could not update operation mode: {str(e)}")

class MQTT:

    """This class implements the MQTT bridge.
    
    Attributes:
        controller (Controller): Connection used to communicate with the device
        connected (bool): Feature used to control the fan speed
        client (Client): Feature used to control the fan speed
        mqtt_cfg (Dict[str,Any]): Feature used to control the fan speed
        loop (AsyncioLoop): Feature used to control the fan speed
    """
  
    ROOT_TOPIC = "/madoka"
    OPERATION_MODE_TOPIC = "operation_mode"
    POWER_STATE_TOPIC = "fan_speed"
    FAN_SPEED_TOPIC = "fan_speed"
    SET_POINT_TOPIC = "set_point"
    AVAILABLE_TOPIC = "available"

    
    def __init__(self,loop,controller: Controller,
                 config: Dict[str,Any]):

        """Initialize the MQTT bridge.
    
        Args:
            loop (AsyncioLoop): Asyncio loop to integrate the MQTT loop with
            controller (Controller): Controller used to manage the device
            mqtt_cfg (Dict[str,Any]): MQTT config 
            
        """
        self.controller:Controller = controller
        self.connected:bool = False
        self.client:mqtt.Client = None
        self.mqtt_cfg = config["mqtt"]
        self.loop = loop
        

    def connect(self):
        """ Connect to the MQTT broker. It returns a future to notify when the 
        connection has been finished.
        """

        id = "madoka_mqtt_" + self.controller.connection.name

        if "id" in self.mqtt_cfg:
            id = self.mqtt_cfg["id"]

        self.client:mqtt.Client = mqtt.Client(client_id = id)
        if "username" in self.mqtt_cfg:
            self.client.username_pw_set(username=self.mqtt_cfg["username"],password=self.mqtt_cfg["password"])
       
        if self.mqtt_cfg["ssl"]:
            self.client.tls_set(ca_certs=mqtt.TLS_CERT_PATH, certfile=None,
                                keyfile=None, cert_reqs=mqtt.ssl.CERT_REQUIRED,
                                tls_version=mqtt.ssl.PROTOCOL_TLSv1_2, ciphers=None)
            self.client.tls_insecure_set(False)
        
        try:

            aioh = AsyncioHelper(self.loop, self.client)
         
            self.client.on_connect = self.on_connect
            self.client.on_message = self.on_message
            self.client.on_disconnect = self.on_disconnect
            self.client.connect(self.mqtt_cfg["host"], port=self.mqtt_cfg["port"])
            self.connect_future = asyncio.get_event_loop().create_future()
            return self.connect_future

        except MQTTException as e:
            logger.error(f"Error in MQTT: {str(e)}")
    
    def start(self):
        """Start the MQTT bridge. Subscribe to the topics"""
        subscribe_topics = []
        for k,v in vars(self.controller).items():
            if isinstance(v,Feature):
                subscribe_topics.append(("/".join([self.get_device_topic(),k,"set"]),0))

        self.client.subscribe(subscribe_topics)

    def stop(self):
        """ Disconnect from the MQTT broker """
        self.client.disconnect()
       
    
    def on_connect(self,client, userdata, flags, rc):
        """ Connection established callback. See paho-mqtt docs for more details. """
        self.connected = rc == 0
        if self.connect_future:
            self.connect_future.set_result(self.connected)
        if self.connected: 
            logger.debug("Connected to MQTT broker")
            self.start()

    def on_disconnect(self,client, userdata, rc):
        """ Connection established callback. See paho-mqtt docs for more details. """
        logger.debug(f"Disconnected from MQTT broker ({rc})")
        asyncio.create_task(self.reconnect())

    async def reconnect(self):
        is_connected = self.client.is_connected        
        while not is_connected:
            try:
                logger.debug("Reconnecting in 60s...")
                await asyncio.sleep(60)
                c = await self.connect()
                await c
                is_connected = c.result()
               
            except CancelledError:
                pass
        
        self.available(self.controller.connection.connection_status == ConnectionStatus.CONNECTED)

    def get_device_topic(self):
        """ Get the customized device topic using the device name and the root topic. 
        Returns:
            str: The device topic in the form /root_topic/device_name"""
        root_topic = self.ROOT_TOPIC 
        if "root_topic" in self.mqtt_cfg:
            root_topic = self.mqtt_cfg["root_topic"]
        normalized_name = self.controller.connection.name
        normalized_name = normalized_name.replace(" ","_")
        normalized_name = normalized_name.replace(":","_")
        normalized_name = normalized_name.replace("/","_")

        return "/".join([root_topic,normalized_name]) 

    def available(self, status:bool):
        """
        Send the status to the availability topic
        Args:
            status (bool): True if available, False otherwise       
        """
        device_topic = self.get_device_topic()
        topic = "/".join([device_topic, self.AVAILABLE_TOPIC])
        
        self.client.publish(topic,"0" if not status else "1")

    def update(self, status:str):
        """
        Send the status to the status topic (JSON payload)
        Args:
            status (str): New status
        """           
        if not self.client.is_connected:
            logger.debug("MQTT broker is not available. Skipping message...")
            device_topic = self.get_device_topic()
            topic = "/".join([device_topic, "state","get"])
            
            self.client.publish(topic,status)
        

    def on_message(self,client, userdata, msg):
        """ Message received callback. See paho-mqtt docs for more details. """
        if msg.topic == "/".join([self.get_device_topic(), self.OPERATION_MODE_TOPIC,"set"]):
            asyncio.get_event_loop().create_task(set_operation_mode(self.controller,msg.payload))
        
        elif msg.topic == "/".join([self.get_device_topic(), self.FAN_SPEED_TOPIC,"set"]):
            asyncio.get_event_loop().create_task(set_fan_speed(self.controller,msg.payload))
        
        elif msg.topic == "/".join([self.get_device_topic(), self.POWER_STATE_TOPIC,"set"]):
            asyncio.get_event_loop().create_task(set_power_state(self.controller,msg.payload))
        
        elif msg.topic == "/".join([self.get_device_topic(), self.SET_POINT_TOPIC,"set"]):
            asyncio.get_event_loop().create_task(set_set_point_state(self.controller,msg.payload))
        
    
def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return asyncio.run(f(*args, **kwargs))
        except KeyboardInterrupt:
            logger.info("User stopped program.")
        except Exception as e:
            logger.error("",e, stack_info=True)
        
    return wrapper

async def periodic_update(interval:int,controller:Controller,mqtt_service:MQTT):
    """ This routine is used to schedule the periodic update of the controller.
    Args:
        interval (int): Number of seconds to wait between updates
        controller (`Controller`): Device controller to be updated
        mqtt_service (`MQTT`): MQTT service to send updates with
    """
    available = True
    while available:
        try:
            await controller.update()
            status = controller.refresh_status()
            mqtt_service.update(json.dumps(status,default=str))
            await asyncio.sleep(interval)
        except CancelledError as e:
                pass
        except ConnectionAbortedError as e:
             mqtt_service.available(False)
             available = False
        except Exception as e:
            logger.error(f"Exception caught : {str(e)}")

@click.command()
@click.pass_context
@click.option('-a', '--address', required=True, type=str, help="Bluetooth MAC address of the thermostat")
@click.option('-c', '--config', required=True, type=click.Path(), help="MQTT config file")
@click.option('-d', '--adapter', required=False, type=str, default="hci0", show_default=True, help="Name of the Bluetooth adapter to be used for the connection")
@click.option('--force-disconnect/--not-force-disconnect', default="True", show_default=True, help="Should disconnect the device to ensure it is recognized (recommended)")
@click.option('-t', '--device-discovery-timeout', type=int, default=5, show_default=True, help = "Timeout for Bluetooth device scan in seconds")
@click.option('-o', '--log-output', required=False,type=click.Path(), help="Path to the log output file")
@click.option('--debug', is_flag=True, help="Enable debug logging")
@click.option('--verbose', is_flag=True, help="Enable versbose logging")
@click.version_option()
@coro
async def run(ctx,verbose,adapter,log_output,debug,address,force_disconnect, device_discovery_timeout, config):
    
        
    madoka = Controller(address, adapter = adapter)
    
    ctx.obj = {}
    ctx.obj["madoka"] = madoka
    ctx.obj["loop"] = asyncio.get_event_loop()   
    ctx.obj["timeout"] = device_discovery_timeout
    ctx.obj["adapter"] = adapter
    ctx.obj["force_disconnect"] = force_disconnect
    with open(config, 'r') as stream:
        yml_config = yaml.safe_load(stream)
        ctx.obj["config"] = yml_config
    
    logging_level = None
    if verbose:
        logging_level = logging.DEBUG if debug else logging.DEBUG
        
    logging_filename = log_output
    logging.basicConfig(level=logging_level,
                        filename = logging_filename)

    
    if force_disconnect:
        await force_device_disconnect(madoka.connection.address)
    discovered_devices = await discover_devices(timeout = ctx.obj["timeout"], adapter = ctx.obj["adapter"])
    mqtt_service = MQTT(asyncio.get_event_loop(),madoka,yml_config)
    update_task = None
    try:
        
        await madoka.start()       
        connect = mqtt_service.connect() 
        
        await connect
        if not connect.result():
              return
        
        mqtt_service.available(True)
        
        update_task = asyncio.create_task(periodic_update(yml_config["daemon"]["update_interval"], madoka, mqtt_service))
    
        asyncio.gather(update_task)
    except CancelledError as e:
        logger.error(e)
    except ConnectionAbortedError as e:
        if mqtt_service.connected:
            mqtt_service.available(False)
        if update_task is not None:
            await update_task.stop()
        mqtt_service.stop()
        await madoka.stop()
        logger.error(e)
    except ConnectionRefusedError as e:
        logger.error("Could not connect to MQTT broker")
    except Exception as e:
        logger.error(e)


if __name__ == "__main__":  
      

    asyncio.run(run())        
    asyncio.get_event_loop().run_forever()

    
   



from setuptools import setup, find_packages

setup(
    name='pymadoka',
    version='0.2.4',
    py_modules=['pymadoka'],
    author = "Manuel Durán",
    author_email = "manuelduran@gmail.com",
    description = ("A library to control Daikin BRC1H (Madoka) thermostats"),
    license = "MIT",
    url = "https://github.com/mduran80/pymadoka",
    keywords = "thermostat homeautomation bluetooth",
    packages=find_packages()+ find_packages(where="./features"),
    install_requires=[
        'click',
        'bleak >=0.10, <0.11',
        'pyyaml',
        'asyncio-mqtt'
    ],
    entry_points='''
        [console_scripts]
        pymadoka=pymadoka.cli:cli
        pymadoka-mqtt=pymadoka.mqtt:run
    ''',
)


import asyncio
import logging
from pymadoka.connection import discover_devices, force_device_disconnect


import click
import time
import json
import threading  

from functools import wraps
from pymadoka.controller import Controller
from pymadoka.features.fanspeed import FanSpeedEnum, FanSpeedStatus
from pymadoka.features.setpoint import SetPointStatus
from pymadoka.features.operationmode import OperationModeStatus, OperationModeEnum
from pymadoka.features.power import PowerStateStatus
from pymadoka.features.clean_filter import ResetCleanFilterTimerStatus

def format_output(format,status):
    print(json.dumps(vars(status),default=str))

import threading
import time

logger = logging.getLogger(__name__)

class LoadingThread(threading.Thread):
    
    def __init__(self):
        super().__init__()
        self.stopevent = threading.Event()

    def stop(self):
        self.stopevent.set()

    def join(self, *args, **kwargs):
        self.stop()
        super().join(*args, **kwargs)

    def run(self):
        spaces = 0  
        while not self.stopevent.is_set():
            print("\b "*spaces+".", end="", flush=True) 
            spaces = spaces+1                          
            time.sleep(0.2)                            
            if (spaces>5):                              
                print("\b \b"*spaces, end="")           
                spaces = 0  
        print("\b"*(spaces+2), end="", flush=True) 
 

def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        async def p(*args,**kwargs):
            
            loading = args[0]["loading"]
            if loading is not None:
                loading.start()

            madoka:Controller = args[0]["madoka"]
            if args[0]["force_disconnect"]:
                await force_device_disconnect(madoka.connection.address)
            await discover_devices(timeout = args[0]["timeout"], adapter = args[0]["adapter"])
            await madoka.start()
            status = await f(*args,**kwargs)
            if loading is not None:
                loading.join()
            format_output(args[0]["format"],status)
            
        try:
            return asyncio.run(p(*args, **kwargs)) 
        except KeyboardInterrupt:
            logger.info("User stopped program.")
        except Exception as e:
            logger.error(str(e))
            if "loading" in args[0]:
                args[0]["loading"].join()
        finally:
            logger.info("Disconnecting...")
   
    return wrapper
@click.group(chain=False)
@click.pass_context
@click.option('-a', '--address', required=True, type=str, help="Bluetooth MAC address of the thermostat")
@click.option('-d', '--adapter', required=False, type=str, default="hci0", show_default=True, help="Name of the Bluetooth adapter to be used for the connection")
@click.option('--force-disconnect/--not-force-disconnect', default="True", show_default=True, help="Should disconnect the device to ensure it is recognized (recommended)")
@click.option('-t', '--device-discovery-timeout', type=int, default=5, show_default=True, help = "Timeout for Bluetooth device scan in seconds")
@click.option('-o', '--log-output', required=False,type=click.Path(), help="Path to the log output file")
@click.option('--debug', is_flag=True, help="Enable debug logging")
@click.option('--verbose', is_flag=True, help="Enable versbose logging")
@click.version_option()


def cli(ctx,verbose,adapter,log_output,debug,address,force_disconnect, device_discovery_timeout):
  
    madoka = Controller(address, adapter = adapter)
    
    ctx.obj = {}
    ctx.obj["madoka"] = madoka
    ctx.obj["loop"] = asyncio.get_event_loop()   
    ctx.obj["format"] = format
    ctx.obj["timeout"] = device_discovery_timeout
    ctx.obj["adapter"] = adapter
    ctx.obj["force_disconnect"] = force_disconnect
 
    loading = LoadingThread()
    logging_level = None
    if verbose:
        logging_level = logging.DEBUG if debug else logging.INFO
    else:
        ctx.obj["loading"] = loading
        
    logging_filename = log_output
    logging.basicConfig(level=logging_level,
                        filename = logging_filename)

    
    
       

@cli.command()
@click.pass_obj
@coro
@click.argument('fan-speed',
              type=(click.Choice(['LOW', 'MID', 'HIGH', 'AUTO'], case_sensitive=True),click.Choice(['LOW', 'MID', 'HIGH', 'AUTO'], case_sensitive=True)))
async def set_fan_speed(obj,fan_speed):
   """Set cooling and heating fan speeds."""
   return await obj["madoka"].fan_speed.update(FanSpeedStatus(FanSpeedEnum(fan_speed[0]), FanSpeedEnum(fan_speed[1])))
  
@cli.command()
@click.pass_obj
@coro
async def get_fan_speed(obj):
    """Get cooling and heating fan speeds."""
    return await obj["madoka"].fan_speed.query()
 

    
@cli.command()
@click.pass_obj
@coro
async def get_operation_mode(obj):
    """Get the operation mode."""
    return await obj["madoka"].operation_mode.query()
  
@cli.command()
@click.pass_obj
@coro
@click.argument('operation-mode',
              type=click.Choice(['FAN', 'DRY', 'AUTO', 'COOL','HEAT','VENTILATION'], case_sensitive=True))
async def set_operation_mode(obj,operation_mode):
    """Set the operation mode."""
    return await obj["madoka"].operation_mode.update(OperationModeStatus(OperationModeEnum(operation_mode)))
    

@cli.command()
@click.pass_obj
@coro
async def get_power_state(obj):
    """Check if the HVAC is turned on."""
    return await obj["madoka"].power_state.query()
  
   
@click.command()
@click.pass_obj
@coro
@click.argument('power-state',
              type=click.Choice(['ON','OFF'], case_sensitive=True))
async def set_power_state(obj,power_state):
    """Turn ON or OFF the HVAC."""
    return await obj["madoka"].set_point.update(PowerStateStatus(power_state == "ON"))
  
        
@cli.command()
@click.pass_obj
@coro
async def get_temperatures(obj):
    """Get temperatures as read by the thermostat."""
    return await obj["madoka"].temperatures.query()
  

@cli.command()
@click.pass_obj
@coro
async def get_set_point(obj):
    """Get target temperatures in Celsius degrees."""
    return await obj["madoka"].set_point.query()

@cli.command()
@click.pass_obj
@coro
@click.argument('set-point',
              type= (click.IntRange(0, 30, clamp=True),click.IntRange(0, 30, clamp=True)))
async def set_set_point(obj,set_point):
    """Set cooling/heating target temperatures in Celsius degrees."""
    return await obj["madoka"].set_point.update(SetPointStatus(set_point[0],set_point[1]))
    
@cli.command()
@click.pass_obj
@coro
async def get_clean_filter_indicator(obj):
    """Get status of the Clean Filter indicator."""
    return await obj["madoka"].clean_filter_indicator.query()
   
@cli.command()
@click.pass_obj
@coro
async def reset_clean_filter_timer(obj):
    """Reset the Clean Filter indicator timer."""
    return await obj["madoka"].reset_clean_filter_timer.update(ResetCleanFilterTimerStatus())
  
@cli.command()
@click.pass_obj
@coro
async def get_status(obj):
    """Get status of all the thermostat features."""
    await obj["madoka"].update()
    return obj["madoka"].refresh_status()
    

@cli.command()
@click.pass_obj
@coro
async def get_info(obj):
    """Get info of the device."""
    return await obj["madoka"].read_info()
  
    

if __name__ == "__main__":  
    asyncio.get_event_loop().run_until_complete(cli())
   



"""This module implements the base class of the features supported by the device
"""

from abc import ABC, abstractmethod
import logging
import json
from typing import Dict

from asyncio.exceptions import CancelledError

from pymadoka.connection import Connection, ConnectionException, ConnectionStatus

logger = logging.getLogger(__name__)

class ParseException(Exception):
     pass

class NotImplementedException(Exception):
     pass

class FeatureStatus(ABC):
    """
    This interface defines the methods used by the Transport to notify the result of the rebuild process.
    """

    """This method must be implemented by subclasses to provide with the list of parameters used by the feature     
    Returns:
        Dict[int,bytearray]: Dictionary of parameter ids and values
    """
    @abstractmethod
    def get_values(self) -> Dict[int,bytearray]:
        pass

    """This method must be implemented by subclasses to provide with the list of parameters used by the feature.

    Returns:
        Dict[int,bytearray]: Dictionary of parameter ids and values
    """
    @abstractmethod
    def set_values(self,values:Dict[int,bytearray]):
        pass

    """Parse the provided data into a dictionary of parameter names and values.

    Once the parameters have been parsed, they are passed to the feature using the method `set_values`
    Args:
        data (bytearray): Data to be parsed    
    Raises:
        ParseException: There is missing data or there is a data mismatch
    """
    def parse(self,data:bytearray):

        if len(data)<4:
            raise ParseException("Not enough bytes to parse")

        if data[0] != len(data):
            raise ParseException("Message size and data size mismatchs")


        # We have already skipped chunk_id(1byte)
        # We process the following data: size(1),cmd_id(3),param_id(1),param_size(1),param_value...
     
        values = {}
        value_size = 0
        i = 4
        while i < len(data):
            if (i+1) >= len(data):
                raise ParseException("Not enough data to parse while processing arguments")
            
            value_id = data[i]
            if data[i+1] == 0xff:
                value_size = 0
            else: 
                value_size = data[i+1]

            if i+1+value_size >= len(data):
                raise ParseException("Not enough data to parse while processing arguments")
            
            value_bytes = data[i+2:i+2+value_size]
            if len(value_bytes) == 0:
                value_bytes = bytes([0x00])
            values[value_id] = value_bytes

            i += 2 + value_size
        
        self.set_values(values)


    """Serialize the status parameters into a bytearray.

    Each parameter is written with the following structure: <param_id><param_contents_size><param_contents>

    Returns:
        bytearray: Data with all the parameter info
    """
    def serialize(self) -> bytearray:

        values = self.get_values()
    
        out = bytearray()

        for k,v in values.items():
            out.append(k)
            out.append(len(v))
            out.extend(v)

        # Special case when no parameters are used

        if len(out) == 0:
            out = bytearray([0x00,0x00])

        return out
            

class Feature(ABC):
    """
    This interface defines the methods used by the features.

    Attributes:
        connection (Connection): Connection to be used to send messages
        status (FeatureStatus): Status 
    """
    def __init__(self, connection: Connection):
        """Inits the feature with the connection.

        Args:
            connection (Connection): Connection to be used to send messages
        """
        self.connection = connection
        self.status = None
        super().__init__()
    
    
    @abstractmethod
    def new_status(self) -> FeatureStatus:
        """This method must be implemented by subclasses to return a new instance of the status used by this feature.

        Returns:
            FeatureStatus: New status instance
        """
        pass

    
    @property
    @abstractmethod
    def query_cmd_id(self) -> int:
        """This method must be implemented by subclasses to return a the id used to query the device feature.

        Returns:
            int: Query status cmd id
        """
        pass


    @property
    @abstractmethod
    def update_cmd_id(self) -> int:
        """This method must be implemented by subclasses to return a the id used to update the device feature.

        Returns:
            int: Update status cmd id
        """
        pass

    async def query(self) -> FeatureStatus:
        """This method is used to query the device for this feature.

        The method waits until the response is received, parses the result and updates the feature state accordingly.

        Returns:
            FeatureStatus: New status
        Raises:
            ConnectionAbortedError: If the connection is not available
            ConnectionException: If an error appeared during message delivery or reception
            Exception: Any other exception raised is bubbled-up
        """

        if self.connection.connection_status == ConnectionStatus.ABORTED:
                raise ConnectionAbortedError(f"Could not send command: connection is not available")

        cmd_id = self.query_cmd_id()
        try:
            
           
             new_status = self.new_status()
             response = await self.connection.send(cmd_id, new_status.serialize())
             await response
             result = response.result()
             logger.debug(f"{self.__class__.__name__} QUERY response received ({len(result)} bytes)")
             new_status.parse(result)
             logger.debug(f"{self.__class__.__name__} status updated, new value:\n{json.dumps(vars(new_status), default = str)}")
             self.status = new_status
             return self.status             
        except CancelledError as e:
            if cmd_id in self.connection.requests:
                if len(self.connection.requests[cmd_id])>0:
                    self.connection.requests[cmd_id].pop()
            if self.connection.connection_status == ConnectionStatus.ABORTED:
                raise ConnectionAbortedError(f"Could not send command: connection is not available")
            elif self.connection.connection_status == ConnectionStatus.CONNECTING:
                pass
            else:
                raise ConnectionException(f"Could not send command: message could not be rebuilt")
        except ConnectionAbortedError as e:
            raise e
        except Exception as e:
            raise e
        

    async def update(self,update_status:FeatureStatus) -> FeatureStatus:
        """This method is used to update the device for this feature.

        The method waits until the response is received, parses the result and updates the feature state accordingly.

        We can assume that if the response was parsed correctly, the command went OK. The response data, algthough parseable, does not
        reflect the actual status of the device. e.g:

        Operation Mode Command Set DRY:
        < ACL Data TX: Handle 73 flags 0x00 dlen 15             #1757 [hci0] 984.889214
        ATT: Write Command (0x52) len 10
        Handle: 0x0205
          Data: 0007004030200101 <---- DRY

        Operation Mode Command Set DRY - Response :
        > ACL Data RX: Handle 73 flags 0x02 dlen 15             #1759 [hci0] 984.951395
        ATT: Handle Value Notification (0x1b) len 10
        Handle: 0x0202
          Data: 0007004030200100 <---- FAN_ONLY

        Please note the last byte as it 

        Args:
            update_status (FeatureStatus): New status to be set
        Returns:
            FeatureStatus: New status
        Raises:
            ConnectionAbortedError: If the connection is not available
            ConnectionException: If an error appeared during message delivery or reception
            Exception: Any other exception raised is bubbled-up
        """

        if self.connection.connection_status == ConnectionStatus.ABORTED:
                raise ConnectionAbortedError(f"Could not send command: connection is not available")

        cmd_id = self.update_cmd_id()
        try:
             response = await self.connection.send(cmd_id, update_status.serialize())
             await response
             result = response.result()
             logger.debug(f"{self.__class__.__name__} UPDATE response received ({len(result)} bytes)")
             response_status = self.new_status()
             response_status.parse(result)
             logger.debug(f"{self.__class__.__name__} status updated, new value:\n{json.dumps(vars(response_status), default = str)}")
             self.status = update_status
             return self.status
        except CancelledError as e:
            if cmd_id in self.connection.requests:
                if len(self.connection.requests[cmd_id])>0:
                    self.connection.requests[cmd_id].pop()
            if self.connection.connection_status == ConnectionStatus.ABORTED:
                raise ConnectionAbortedError(f"Could not send command: connection is not available")
            elif self.connection.connection_status == ConnectionStatus.CONNECTING:
                pass
            else:
                raise ConnectionException(f"Could not send command: message could not be rebuilt")
        except ConnectionAbortedError as e:
            raise e
        except Exception as e:
            raise e
       
